#!/usr/bin/gawk -f
#
# toptalkers.awk -- per-user traffic analysis for D-Link Firewall
# Placed in the public domain by the author, <mikael.olsson@clavister.com>.
#
# Version: 1.2
#
# 1.0: First public release
# 1.1: Added user name display ("u" switch)
# 1.2: Added data-per-port display ("p" switch)
#
# NOTE: requires gawk v3.1 or newer
#


function usage() {
  print "usage:";
  print "  toptalkers <interface> [<viewnum> [<whichstats>]]";
  print "";
  print "    <interface>  -- traffic to/from users behind this interface is counted";
  print "    <viewnum>    -- max. number of users to display";
  print "    <whichstats> -- one or more of the letters 'crsdu'";
  print "                    c: connection count to/from user";
  print "                    r: data received by user";
  print "                    s: data received by user";
  print "                    d: data received + sent by user";
  print "                    u: user names rather than IP addresses (when available)";
  print "                    p: data transfered per destination port";
  print "";
}



#
# BEGIN: 
#   Parse commandline arguments.
#   Initialize stuff.
#

BEGIN {

  if(ARGC<2 || ! ( ARGV[1] ~ /^[a-zA-Z(]/ ) )
  {
    usage();
    exit(0);
  }

  # The interface to stat
  strIface = ARGV[1];
  
  # How many entries to view -- default 10
  if(ARGV[2]!="")
    nEntries=int(ARGV[2]);
  else
    nEntries=10;

  # What stats to show -- default ALL
  bConn=1; bData=1; bSent=1; bRecv=1; bUsernames=1; bPortstats=1;
  if(ARGV[3] != "")
  {
    bConn=0; bData=0; bSent=0; bRecv=0; bUsernames=0; bPortstats=0;
    if(ARGV[3] ~ /c/)
      bConn=1;
    if(ARGV[3] ~ /d/)
      bData=1;
    if(ARGV[3] ~ /r/)
      bRecv=1;
    if(ARGV[3] ~ /s/)
      bSent=1;
    if(ARGV[3] ~ /u/)
      bUsernames=1;
    if(ARGV[3] ~ /p/)
      bPortstats=1;
  }

  # Make awk ignore our argument list -- otherwise it'll attempt to parse it as file names
  ARGC=1;

  # Windows users can't just use "sort", or they might accidentally use windows' own sort utility. That won't do.
  sortcmd=ENVIRON["sortcmd"];
  if(length(sortcmd)<2)
    sortcmd="sort";
}

# scrubproto(): fwlogqry displays protocol as "6(tcp)", which looks a bit messy here, so we reformat it
function scrubproto(str,   a) {
  if(match(str, "^[0-9]+\\(([^)]+)",a) > 0 )
    return tolower(a[1]);
  return tolower(str);
}


#
# We work with "conn=close" events that involve either 
# the interface we're interested in as "recvif=" or "destif".
# For the latter, we have to swap the recv/sent transfer amounts.
#

/conn=close/ {
  if(match($0, "recvif=" strIface " .*srcip=([^ ]+).*origsent=([0-9]+).*termsent=([0-9]+)", a)>0) 
  {
    if(bUsernames && match($0, "srcusername=(\"[^\"]+\"|[^ ]+)", b)>0)
      user=b[1];
    else
      user=a[1];
    num[user]++;
    sent[user]+=int(a[2]);
    recv[user]+=int(a[3]);

    if(bPortstats)
    {
      if(match($0, "proto=([^ ]+).*destport=([^ ]+)", b)>0)
        foo = b[2] "/" scrubproto(b[1]);
      else if(match($0, "proto=([^ ]+)", b)>0)
        foo = scrubproto(b[1]);
      else 
        foo = "?";
      portsent_recvif[foo]+=int(a[2]);
      portrecv_recvif[foo]+=int(a[3]);
    }
  }
  if(match($0, "destif=" strIface " .*destip=([^ ]+).*origsent=([0-9]+).*termsent=([0-9]+)", a)>0) 
  {
    if(bUsernames && match($0, "destusername=(\"[^\"]+\"|[^ ]+)", b)>0)
      user=b[1];
    else
      user=a[1];
    num[user]++;
    sent[user]+=int(a[3]);
    recv[user]+=int(a[2]);

    if(bPortstats)
    {
      if(match($0, "proto=([^ ]+).*destport=([^ ]+)", b)>0)
        foo = b[2] "/" scrubproto(b[1]);
      else if(match($0, "proto=([^ ]+)", b)>0)
        foo = scrubproto(b[1]);
      else 
        foo = "?";
      portsent_destif[foo]+=int(a[3]);
      portrecv_destif[foo]+=int(a[2]);
    }
  }
}



#
# Utility function: fmtnum()
# Outputs a number in groups of three, in a 11-char-wide field
#

function fmtnum(num) {
  num = sprintf("%12u", num);
  return substr(num, 1,3) " " substr(num,4,3) " " substr(num,7,3) " " substr(num,10)
}



#
# END: When we're done collecting data, we output the stats
#      that we were asked to display.
#

END {

  #
  # Top connection statistics
  #

  if(bConn) 
  {
    print ""
    print "  Top " nEntries " talkers -- connections"
    print "  -----------------------------"
  
    cmd = sortcmd " -r | head -" nEntries;
    for(i in num)
      printf("  %s  %s\n", fmtnum(num[i]), i) | cmd;

    close(cmd);
  }

  #
  # Top data (recv+sent) statistics. 
  #

  if(bData) 
  {
    print ""
    print "  Top " nEntries " talkers -- data"
    print "  ---------------------------"
    
    cmd = sortcmd " -r | head -" nEntries;
    for(i in sent)
      printf("  %sK %-15s (%sK sent + %sK recvd)\n", fmtnum((sent[i]+recv[i])/1024), i, fmtnum(sent[i]/1024), fmtnum(recv[i]/1024)) | cmd;

    close(cmd);
  }

  #
  # Top sent statistics
  #

  if(bSent) 
  {
    print ""
    print "  Top " nEntries " talkers -- sent data"
    print "  ---------------------------"
    
    cmd = sortcmd " -r | head -" nEntries;
    for(i in sent)
      printf("  %sK %s\n", fmtnum(sent[i]/1024), i) | cmd;

    close(cmd);
  }

  #
  # Top received statistics
  #
  
  if(bRecv) 
  {
    print ""
    print "  Top " nEntries " talkers -- recvd data"
    print "  ----------------------------"
    
    cmd = sortcmd " -r | head -" nEntries;
    for(i in recv)
      printf("  %sK %s\n", fmtnum(recv[i]/1024), i) | cmd;

    close(cmd);
  }

  #
  # Top per-port statistics, destif=<myiface>
  #
  
  if(bPortstats) 
  {
    print "";
    print "  Top " nEntries " ports, destif=" strIface "         Note: 'sent'='sent by server'";
    print "  ----------------------------";
    
    cmd = sortcmd " -r | head -" nEntries;
    for(i in portsent_destif)
      printf("  %sK %-15s (%sK sent + %sK recvd)\n", fmtnum((portsent_destif[i]+portrecv_destif[i])/1024), i, fmtnum(portsent_destif[i]/1024), fmtnum(portrecv_destif[i]/1024)) | cmd;

    close(cmd);
  }

  #
  # Top per-port statistics, recvif=<myiface>
  #
  
  if(bPortstats) 
  {
    print "";
    print "  Top " nEntries " ports, recvif=" strIface "         Note: 'sent'='sent by client'";;
    print "  ----------------------------";
    
    cmd = sortcmd " -r | head -" nEntries;
    for(i in portsent_recvif)
      printf("  %sK %-15s (%sK sent + %sK recvd)\n", fmtnum((portsent_recvif[i]+portrecv_recvif[i])/1024), i, fmtnum(portsent_recvif[i]/1024), fmtnum(portrecv_recvif[i]/1024)) | cmd;

    close(cmd);
  }

}


